import { encodeBase32LowerCaseNoPadding, encodeHexLowerCase } from "@oslojs/encoding";
import { sha256 } from "@oslojs/crypto/sha2";
import type { Session, SessionValidationResult, User } from ".";

interface ValidateResponse {
	id: string;
	user_id: string;
	expires_at: number;
	name: string;
	avatar_hash: string;
}

export function generateSessionToken(): string {
	const bytes = new Uint8Array(40);
	crypto.getRandomValues(bytes);
	return encodeBase32LowerCaseNoPadding(bytes);
}

export async function createSession(
	token: string,
	userId: string,
	db: D1Database,
): Promise<Session> {
	const sessionId = encodeHexLowerCase(sha256(new TextEncoder().encode(token)));
	const session: Session = {
		id: sessionId,
		userId,
		expiresAt: new Date(Date.now() + 1000 * 60 * 60 * 24 * 7), // 7 days
	};

	await db
		.prepare("INSERT INTO session (id, user_id, expires_at) VALUES (?, ?, ?)")
		.bind(session.id, session.userId, Math.floor(session.expiresAt.getTime() / 1000))
		.run();

	return session;
}

export async function validateSessionToken(
	token: string,
	db: D1Database,
): Promise<SessionValidationResult> {
	const sessionId = encodeHexLowerCase(sha256(new TextEncoder().encode(token)));

	const row = await db
		.prepare(
			"SELECT session.id, session.user_id, session.expires_at, user.name, user.avatar_hash FROM session INNER JOIN user ON user.id = session.user_id WHERE session.id = ?",
		)
		.bind(sessionId)
		.first<ValidateResponse>();

	if (row === null) return { session: null, user: null };

	const session: Session = {
		id: row.id,
		userId: row.user_id,
		expiresAt: new Date(row.expires_at * 1000),
	};

	const user: User = {
		id: row.user_id,
		name: row.name,
		avatarHash: row.avatar_hash,
	};

	if (Date.now() >= session.expiresAt.getTime()) {
		await invalidateSession(sessionId, db);
		return { session: null, user: null };
	}

	// if less than 3 days left
	if (Date.now() >= session.expiresAt.getTime() - 1000 * 60 * 60 * 24 * 3) {
		session.expiresAt = new Date(Date.now() + 1000 * 60 * 60 * 24 * 7); // 7 days

		await db
			.prepare("UPDATE session SET expires_at = ? WHERE id = ?")
			.bind(Math.floor(session.expiresAt.getTime() / 1000), session.id)
			.run();
	}

	return { session, user };
}

export async function invalidateSession(sessionId: string, db: D1Database): Promise<void> {
	await db.prepare("DELETE FROM session WHERE id = ?").bind(sessionId).run();
}
