import { error } from "@sveltejs/kit";
import type { CheckResponse, ClientGameData, GameOptions } from "./types";

const metroUrl =
	"https://data.grandlyon.com/geoserver/sytral/ows?SERVICE=WFS&VERSION=2.0.0&request=GetFeature&typename=sytral:tcl_sytral.tcllignemf_2_0_0&outputFormat=application/json&SRSNAME=EPSG:4171&sortBy=gid";
const tramUrl =
	"https://data.grandlyon.com/geoserver/sytral/ows?SERVICE=WFS&VERSION=2.0.0&request=GetFeature&typename=sytral:tcl_sytral.tcllignetram_2_0_0&outputFormat=application/json&SRSNAME=EPSG:4171&sortBy=gid";
const stopsUrl =
	"https://data.grandlyon.com/geoserver/sytral/ows?SERVICE=WFS&VERSION=2.0.0&request=GetFeature&typename=sytral:tcl_sytral.tclarret&outputFormat=application/json&SRSNAME=EPSG:4171&sortBy=gid";

let lazyMetro: [GeoJSON.Feature, string][] | null = null;
let lazyTram: [GeoJSON.Feature, string][] | null = null;
let lazyStops: GeoJSON.Feature[] | null = null;

export async function createGame(
	userId: string | null,
	mode: string,
	stopsType: string,
	stops: GeoJSON.Feature[],
	db: D1Database,
): Promise<ClientGameData> {
	const uuid = crypto.randomUUID();
	const stopNames = stops.map((s) => s.properties!.nom);

	const gameSql = db
		.prepare(
			"INSERT INTO game (id, user_id, mode, stops_type, time, total_score) VALUES (?, ?, ?, ?, unixepoch('now'), 0)",
		)
		.bind(uuid, userId, mode, stopsType);

	const roundSql = db.prepare(
		"INSERT INTO round (game_id, points, distance, stop_name) VALUES (?, -1, -1, ?)",
	);

	await db.batch([gameSql, ...stopNames.map((s) => roundSql.bind(uuid, s))]);

	return {
		gameId: uuid,
		stopNames,
	};
}

export async function saveScore(
	uuid: string,
	stopName: string,
	distanceCalc: (latlng: [number, number]) => number,
	scoreCalc: (dist: number) => number,
	db: D1Database,
): Promise<CheckResponse | null> {
	const row = await db
		.prepare(
			"SELECT round.id, round.stop_name, round.points, game.stops_type FROM round INNER JOIN game ON game.id = round.game_id WHERE round.game_id = ? AND round.stop_name = ?",
		)
		.bind(uuid, stopName)
		.first();

	if (row === null || row.points !== -1) return null;

	const type = row.stops_type as "metro" | "tram";
	const stops = (await getMergedStops(type)).filter((f) => f.properties?.nom === stopName);

	if (stops.length === 0) return null;

	const coords = (stops[0].geometry as GeoJSON.Point).coordinates;
	const latlng: [number, number] = [coords[1], coords[0]];

	const distance = Math.round(distanceCalc(latlng));
	const score = Math.round(scoreCalc(distance));

	await db.batch([
		db
			.prepare("UPDATE round SET points = ?, distance = ? WHERE id = ?")
			.bind(score, distance, row.id),
		db.prepare("UPDATE game SET total_score = total_score + ? WHERE id = ?").bind(score, uuid),
	]);

	return {
		distance,
		score,
		solution: latlng,
	};
}

export async function getMetro(): Promise<[GeoJSON.Feature, string][]> {
	if (!lazyMetro) {
		const metro: GeoJSON.FeatureCollection = await fetch(metroUrl).then((r) => r.json());

		lazyMetro = metro.features.map((f) => {
			const components = f.properties!.couleur!.split(" ");
			return [f, `rgb(${components.join(",")})`];
		});
	}

	return lazyMetro;
}

export async function getTram(): Promise<[GeoJSON.Feature, string][]> {
	if (!lazyTram) {
		const tram: GeoJSON.FeatureCollection = await fetch(tramUrl).then((r) => r.json());

		lazyTram = tram.features.map((f) => {
			const components = f.properties!.couleur!.split(" ");
			return [f, `rgb(${components.join(",")})`];
		});
	}

	return lazyTram;
}

async function getLineCodes(stopsType: "metro" | "tram"): Promise<Set<string>> {
	let lines;

	switch (stopsType) {
		case "metro":
			lines = await getMetro();
			break;
		case "tram":
			lines = await getTram();
			break;
	}

	return new Set<string>(lines.map(([f]) => f.properties!.code_ligne));
}

async function getStops(): Promise<GeoJSON.Feature[]> {
	if (!lazyStops) {
		const stops: GeoJSON.FeatureCollection = await fetch(stopsUrl).then((r) => r.json());
		lazyStops = stops.features;
	}

	return lazyStops;
}

export async function getMergedStops(stopsType: "metro" | "tram"): Promise<GeoJSON.Feature[]> {
	const stops = await getStops();
	const lineCodes = await getLineCodes(stopsType);

	const crossingStops = stops.filter((f) => {
		if (f.properties?.desserte) {
			return f.properties.desserte
				.split(",")
				.map((d: string) => d.split(":")[0])
				.some((d: string) => lineCodes.has(d));
		} else {
			return false;
		}
	});

	// merge coordinates of the stops
	const mergedStops = Array.from(new Set(crossingStops.map((f) => f.properties?.nom))).map(
		(nom) => {
			const matching = crossingStops.filter((f) => f.properties?.nom === nom);

			const longs = matching.map((f) => (f.geometry as GeoJSON.Point).coordinates[0]);
			const lonAvg = longs.reduce((a, b) => a + b) / longs.length;

			const lats = matching.map((f) => (f.geometry as GeoJSON.Point).coordinates[1]);
			const latAvg = lats.reduce((a, b) => a + b) / lats.length;

			const desserte = matching.map((f) => f.properties?.desserte).join(",");

			const feature: GeoJSON.Feature = {
				...matching[0],
				geometry: { type: "Point", coordinates: [lonAvg, latAvg] },
				properties: { ...matching[0].properties, desserte },
			};

			return feature;
		},
	);

	return mergedStops;
}

export function parseOptions(url: URL): GameOptions {
	const mode = url.searchParams.get("mode");
	if (mode !== "easy" && mode !== "hard" && mode !== "extreme demon ultra miguel") {
		error(400, "gamemode is invalid");
	}

	const stopsType = url.searchParams.get("stops_type");
	if (stopsType !== "metro" && stopsType !== "tram") {
		error(400, "stops_type is invalid");
	}

	return { mode, stopsType };
}

// https://underscorejs.org/docs/modules/sample.html
export function sample<T>(arr: T[], n: number): T[] {
	const last = arr.length - 1;
	for (let index = 0; index < n; index++) {
		const rand = index + Math.floor(Math.random() * (last - index + 1));
		const temp = arr[index];
		arr[index] = arr[rand];
		arr[rand] = temp;
	}
	return arr.slice(0, n);
}
